class ExceptionCouldNotDownloadData(Exception):
    """
    Query raised when no frequency data is available for the spotter
    """
    pass


class ExceptionNoDataForVariable(Exception):
    """
    Query raised when no frequency data is available for the spotter
    """
    pass
